var CheckBoxes = [];
var Circleables = [];

/**
 * Adds Click-Listeners to Checkboxes and registers them in the CheckBoxes array
 */
function addCBListeners() {
    let cbArr = document.querySelectorAll("div.cb");
    for (let i = 0; i < cbArr.length; i++) {
        let cbElem = cbArr[i];
        
        cbElem.setAttribute("onclick", "cbOnClick(this);");

        if (CheckBoxes[cbElem.getAttribute("name")] !== undefined) {
            console.error(
                "function 'addCBListeners', page '" + window.location.href + "': Faulty HTML detected! Checkbox '" + cbElem.getAttribute("name") + "' exists multiple times!"
            );
            alert("A critical error occured creating the functionality for the checkboxes. Sadly, this error will prevent the site from working correctly. Check the console for a more detailed info.");
            return;
        }

        CheckBoxes[cbElem.getAttribute("name").toString()] = {
            isSelected: false,
            isDiscarded: false,
            element: cbElem,
            name: cbElem.getAttribute("name")
        };
    }

    let circleableArr = document.querySelectorAll("div.circleable");
    for (let i = 0; i < circleableArr.length; i++) {
        let cElem = circleableArr[i];
        
        cElem.setAttribute("onclick", "circleableOnClick(this);");

        if (Circleables[cElem.getAttribute("name").toString()] !== undefined) {
            console.error(
                "function 'addCBListeners', page '" + window.location.href + "': Faulty HTML detected! Checkbox '" + cElem.getAttribute("name") + "' exists multiple times!"
            );
            alert("A critical error occured creating the functionality for the checkboxes. Sadly, this error will prevent the site from working correctly. Check the console for a more detailed info.");
            return;
        }

        Circleables[cElem.getAttribute("name")] = {
            isSelected: false,
            isDiscarded: false,
            element: cElem,
            name: cElem.getAttribute("name")
        };
    }
}

/**
 * Handles click events for custom phasmophobia-style checkboxes with three states
 * @param {htmlElement} elem Checkbox element ("div.cb" with unqiue attribute "name")
 */
function cbOnClick(elem) {
    let clickedName = elem.getAttribute("name");
    let clickedElem = CheckBoxes[clickedName];
    let isTriState = elem.classList.contains("tristate");

    if (!clickedElem.isSelected && !clickedElem.isDiscarded) {
        // First Click: Nothing to selected
        setCBState(clickedElem, true, false);

    } else if (clickedElem.isSelected && !clickedElem.isDiscarded) {
        // Second Click: Selected to crossed out or nothing
        if (isTriState)
            setCBState(clickedElem, false, true);
        else
            setCBState(clickedElem, false, false);

    } else if (!clickedElem.isSelected && clickedElem.isDiscarded) {
        // Third Click: Crossed to nothing
        setCBState(clickedElem, false, false);

    } else {
        // Else: Illegal state - correct to nothing
        setCBState(clickedElem, false, false);

    }
}

/**
 * Handles click events for custom circle- and strikeables with three states
 * @param {htmlElement} elem Circleables element ("div.circleable" with unqiue attribute "name")
 */
function circleableOnClick(elem) {
    let clickedName = elem.getAttribute("name");
    let clickedElem = Circleables[clickedName];
    let isTriState = elem.classList.contains("tristate");

    if (!clickedElem.isSelected && !clickedElem.isDiscarded) {
        // First Click: Nothing to selected
        setCState(clickedElem, true, false);

    } else if (clickedElem.isSelected && !clickedElem.isDiscarded) {
        // Second Click: Selected to crossed out or nothing
        if (isTriState)
            setCState(clickedElem, false, true);
        else
            setCState(clickedElem, false, false);

    } else if (!clickedElem.isSelected && clickedElem.isDiscarded) {
        // Third Click: Crossed to nothing
        setCState(clickedElem, false, false);

    } else {
        // Else: Illegal state - correct to nothing
        setCState(clickedElem, false, false);

    }
}

/**
 * Sets the HTML and the CheckBoxes array values such that the document's state
 * corresponds to the supplied values.
 * @param {object} checkBoxesElem Element from the CheckBoxes array, that is to be set
 * @param {boolean} isSelected if true, the checkbox will contain an "X"
 * @param {boolean} isDiscarded if true, the label of the checkbox will be strikethrough
 */
function setCBState(checkBoxesElem, isSelected = false, isDiscarded = false) {

    // Set document properties
    let boxElemArr = checkBoxesElem.element.querySelectorAll("div.cb-box");
    let textElemArr = checkBoxesElem.element.querySelectorAll("p");

    if (boxElemArr.length != 1 || textElemArr.length != 1) {
        console.error(
            "function 'setCBState', page '" + window.location.href + "': CheckBox '" + checkBoxesElem.name + "' has faulty HTML structure: \n" + checkBoxesElem.element.innerHTML 
          + "\nExpected was two elements, one empty div with class cb-box, one p with attribute text and text inside the p-tag."
        );
        alert("An error occured. Please reload the page, it might resolve the issue. Check the console for more detailed error info.");
        return;
    }

    let boxElem = boxElemArr[0];
    let textElem = textElemArr[0];

    if (isSelected)
        boxElem.setAttribute("crossed", null);
    else
        boxElem.removeAttribute("crossed");

    if (isDiscarded)
        textElem.setAttribute("strikethrough", null)
    else
        textElem.removeAttribute("strikethrough");

    // Set array properties

    CheckBoxes[checkBoxesElem.name].isSelected = isSelected;
    CheckBoxes[checkBoxesElem.name].isDiscarded = isDiscarded;

}

/**
 * Sets the HTML and the CheckBoxes array values such that the document's state
 * corresponds to the supplied values.
 * @param {object} checkBoxesElem Element from the CheckBoxes array, that is to be set
 * @param {boolean} isSelected if true, the checkbox will contain an "X"
 * @param {boolean} isDiscarded if true, the label of the checkbox will be strikethrough
 */
function setCState(circleableBoxesElem, isSelected = false, isDiscarded = false) {

    // Deselect others
    let selecteds = document.querySelectorAll("div.circleable[crossed]");
    if (isSelected)
        for (let i = 0; i < selecteds.length; i++)
            setCState(Circleables[selecteds[i].getAttribute("name")], false, false);

    // Set document properties
    let textElemArr = circleableBoxesElem.element.querySelectorAll("p");

    if (textElemArr.length != 1) {
        console.error(
            "function 'setCState', page '" + window.location.href + "': Circleable '" + circleableBoxesElem.name + "' has faulty HTML structure: \n" + circleableBoxesElem.element.innerHTML 
          + "\nExpected was two elements, one empty div with class cb-box, one p with attribute text and text inside the p-tag."
        );
        alert("An error occured. Please reload the page, it might resolve the issue. Check the console for more detailed error info.");
        return;
    }

    let textElem = textElemArr[0];

    if (isSelected)
        circleableBoxesElem.element.setAttribute("crossed", null);
    else
        circleableBoxesElem.element.removeAttribute("crossed");

    if (isDiscarded)
        textElem.setAttribute("strikethrough", null)
    else
        textElem.removeAttribute("strikethrough");

    // Set array properties

    Circleables[circleableBoxesElem.name].isSelected = isSelected;
    Circleables[circleableBoxesElem.name].isDiscarded = isDiscarded;

}

addCBListeners();