/**
 * Ghost names
 */
const ghostNames = {
    spirit:         "Spirit",
    wraith:         "Wraith",
    phantom:        "Phantom",
    poltergeist:    "Poltergeist",
    banshee:        "Banshee",
    jinn:           "Jinn",
    mare:           "Mare",
    revenant:       "Revenant",
    shade:          "Shade",
    demon:          "Demon",
    yurei:          "Yurei",
    oni:            "Oni",
    yokai:          "Yokai",
    hantu:          "Hantu",
    goryo:          "Goryo",
    myling:         "Myling",
    onryo:          "Onryo",
    twins:          "The Twins",
    raiju:          "Raiju",
    obake:          "Obake",
    mimic:          "The Mimic",
    moroi:          "Moroi",
    deogen:         "Deogen",
    thaye:          "Thaye"
}

/**
 * Creates the ghost object based on the given name.
 * @param {string} ghostName The name of the ghost refering to constant object ghostNames.
 * @returns Object of the ghost
 */
function createGhost(ghostName = ghostNames.spirit) {
    let newGhost = {};
    
    // 01 - Journal Info

    newGhost.journalInfo = {};

        // 01.01 - Journal ID

    newGhost.journalInfo.ID = 
        ghostName === ghostNames.spirit         ? "01"
    :   ghostName === ghostNames.wraith         ? "02"
    :   ghostName === ghostNames.phantom        ? "03"
    :   ghostName === ghostNames.poltergeist    ? "04"
    :   ghostName === ghostNames.banshee        ? "05"
    :   ghostName === ghostNames.jinn           ? "06"
    :   ghostName === ghostNames.mare           ? "07"
    :   ghostName === ghostNames.revenant       ? "08"
    :   ghostName === ghostNames.shade          ? "09"
    :   ghostName === ghostNames.demon          ? "10"
    :   ghostName === ghostNames.yurei          ? "11"
    :   ghostName === ghostNames.oni            ? "12"
    :   ghostName === ghostNames.yokai          ? "13"
    :   ghostName === ghostNames.hantu          ? "14"
    :   ghostName === ghostNames.goryo          ? "15"
    :   ghostName === ghostNames.myling         ? "16"
    :   ghostName === ghostNames.onryo          ? "17"
    :   ghostName === ghostNames.twins          ? "18"
    :   ghostName === ghostNames.raiju          ? "19"
    :   ghostName === ghostNames.obake          ? "20"
    :   ghostName === ghostNames.mimic          ? "21"
    :   ghostName === ghostNames.moroi          ? "22"
    :   ghostName === ghostNames.deogen         ? "23"
    :   ghostName === ghostNames.thaye          ? "24"
    :                                             "00";

        // 01.02 - Ghost Name

    newGhost.journalInfo.name = ghostName;

        // 01.03 - Evidences

            // 01.03.01 - Normal Evidences

    newGhost.journalInfo.evidences = [];

                // 01.03.01.01 - EMF 5

    if ([
        ghostNames.spirit, 
        ghostNames.wraith,
        ghostNames.jinn,
        ghostNames.shade,
        ghostNames.oni,
        ghostNames.goryo,
        ghostNames.myling,
        ghostNames.twins,
        ghostNames.raiju,
        ghostNames.obake
    ].includes(ghostName))
        newGhost.journalInfo.evidences[newGhost.journalInfo.evidences.length] = evidences.emf;

                // 01.03.01.02 - Ultraviolet

    if ([
        ghostNames.phantom,
        ghostNames.poltergeist,
        ghostNames.banshee,
        ghostNames.jinn,
        ghostNames.demon,
        ghostNames.hantu,
        ghostNames.goryo,
        ghostNames.myling,
        ghostNames.obake,
        ghostNames.mimic
    ].includes(ghostName))
        newGhost.journalInfo.evidences[newGhost.journalInfo.evidences.length] = evidences.uv;
            
                // 01.03.01.03 - Ghost Writing

    if ([
        ghostNames.spirit,
        ghostNames.poltergeist,
        ghostNames.mare,
        ghostNames.revenant,
        ghostNames.shade,
        ghostNames.demon,
        ghostNames.myling,
        ghostNames.moroi,
        ghostNames.deogen,
        ghostNames.thaye
    ].includes(ghostName))
        newGhost.journalInfo.evidences[newGhost.journalInfo.evidences.length] = evidences.writing;
                        
                // 01.03.01.04 - Freezing Temperatures

    if ([
        ghostNames.jinn,
        ghostNames.revenant,
        ghostNames.shade,
        ghostNames.demon,
        ghostNames.yurei,
        ghostNames.oni,
        ghostNames.hantu,
        ghostNames.onryo,
        ghostNames.twins,
        ghostNames.mimic,
        ghostNames.moroi
    ].includes(ghostName))
        newGhost.journalInfo.evidences[newGhost.journalInfo.evidences.length] = evidences.freezing;
                           
                // 01.03.01.05 - D.O.T.S projector

    if ([
        ghostNames.wraith,
        ghostNames.phantom,
        ghostNames.banshee,
        ghostNames.yurei,
        ghostNames.oni,
        ghostNames.yokai,
        ghostNames.goryo,
        ghostNames.raiju,
        ghostNames.deogen,
        ghostNames.thaye
    ].includes(ghostName))
        newGhost.journalInfo.evidences[newGhost.journalInfo.evidences.length] = evidences.dots;
                                   
                // 01.03.01.06 - Ghost orb

    if ([
        ghostNames.banshee,
        ghostNames.mare,
        ghostNames.revenant,
        ghostNames.yurei,
        ghostNames.yokai,
        ghostNames.hantu,
        ghostNames.onryo,
        ghostNames.raiju,
        ghostNames.obake,
        ghostNames.thaye
    ].includes(ghostName))
        newGhost.journalInfo.evidences[newGhost.journalInfo.evidences.length] = evidences.ghostorb;
                                                
                // 01.03.01.07 - Spirit Box

    if ([
        ghostNames.spirit,
        ghostNames.wraith,
        ghostNames.phantom,
        ghostNames.poltergeist,
        ghostNames.mare,
        ghostNames.yokai,
        ghostNames.onryo,
        ghostNames.twins,
        ghostNames.mimic,
        ghostNames.moroi,
        ghostNames.deogen
    ].includes(ghostName))
        newGhost.journalInfo.evidences[newGhost.journalInfo.evidences.length] = evidences.spiritbox;
                                  
            // 01.03.02 - Forced Evidence

    newGhost.journalInfo.forcedEvidence =
        ghostName === ghostNames.hantu  ? evidences.freezing
    :   ghostName === ghostNames.goryo  ? evidences.dots
    :   ghostName === ghostNames.obake  ? evidences.uv
    :   ghostName === ghostNames.mimic  ? evidences.ghostorb
    :   ghostName === ghostNames.moroi  ? evidences.spiritbox
    :   ghostName === ghostNames.deogen ? evidences.spiritbox
    :                                     null;

            // 01.03.03 - Hidden Evidence

    newGhost.journalInfo.hiddenEvidence =
        ghostName === ghostNames.mimic ? evidences.ghostorb
    :                                    null;

    // 02 - Abilities

    newGhost.abilities = {};

        // 02.01 - Teleports to player with EMF 2

    newGhost.abilities.teleportsToPlayer = ghostName === ghostNames.wraith;

        // 02.02 - Wanders to player with EMF 2

    newGhost.abilities.wandersToPlayer = ghostName === ghostNames.phantom;

        // 02.03 - Polter Explosion

    newGhost.abilities.polterExplosion = ghostName === ghostNames.poltergeist;

        // 02.04 - Drains Sanity by 25% if breaker is on + Breaker-EMF

    newGhost.abilities.sanityDrainBreaker = ghostName === ghostNames.jinn;

        // 02.05 - Lights Off - Instantly turns off light that is switched on

    newGhost.abilities.lightsOff = ghostName === ghostNames.mare;

        // 02.06 - Demon Hunt - Instantly starts a hunt, regarding of sanity

    newGhost.abilities.demonHunt = ghostName === ghostNames.demon;

        // 02.07 - Drains Sanity by 15% if near player and close a door in the room

    newGhost.abilities.sanityDrainDoor = ghostName === ghostNames.yurei;

        // 02.08 - Twin Interaction

    newGhost.abilities.twinInteraction = ghostName === ghostNames.twins;

        // 02.09 - Faster Fingerprint disapperance

    newGhost.abilities.fingerprintsDisappearFaster = ghostName === ghostNames.obake;

        // 02.10 - Moroi Curse

    newGhost.abilities.moroiCurse = ghostName === ghostNames.moroi;

        // 02.11 - Double Door Touch

    newGhost.abilities.doubleDoorTouch = ghostName === ghostNames.yurei;

        // 02.12 - Wander to target

    newGhost.abilities.wanderToTarget = ghostName === ghostNames.banshee;

    // 03 - Traits

    newGhost.traits = {};

        // 03.01 - Active traits (Additional behaviour)

    newGhost.traits.active = {};

            // 03.01.01 - The ghost has a target player

    newGhost.traits.active.hasTargetPlayer = ghostName === ghostNames.banshee;

            // 03.01.02 - Wanders away from lights

    newGhost.traits.active.wandersAwayIfLightsAreOn = ghostName === ghostNames.mare;

            // 03.01.03 - Starts hunt after three blown out fires

    newGhost.traits.active.huntsIfThreeFiresAreBlownOut = ghostName === ghostNames.onryo;

            // 03.01.04 - Can hunt after blowing out fire

    newGhost.traits.active.canHuntIfFireIsBlownOut = ghostName === ghostNames.onryo;

            // 03.01.05 - Changes ghost traits every 30-120 seconds

    newGhost.traits.active.changesGhostTraits = ghostName === ghostNames.mimic;

            // 03.01.06 - Ghost ages and becomes less active

    newGhost.traits.active.canAge = ghostName === ghostNames.thaye;

            // 03.01.07 - Randomly Walks to Target

    newGhost.traits.active.randomlyWandersToTargetPlayer = ghostName === ghostNames.banshee;

            // 03.01.08 - can close front dorr if no player is in the house

    newGhost.traits.active.canCloseFrontDoor = ghostName === ghostNames.yurei;

        // 03.02 - Passive traits (Altered behaviour)

    newGhost.traits.passive = {};

            // 03.02.01 - Cant walk through salt

    newGhost.traits.passive.cantWalkThroughSalt = ghostName === ghostNames.wraith;

            // 03.02.02 - Sanity Drain by 2% for each thrown item

    newGhost.traits.passive.throwsDrainSanity = ghostName === ghostNames.poltergeist;

            // 03.02.03 - More singing ghost events

    newGhost.traits.passive.moreSingingGhostEvents = ghostName === ghostNames.banshee;

            // 03.02.04 - Can't disable fuse box

    newGhost.traits.passive.cantDisableFuseBox = ghostName === ghostNames.jinn;

            // 03.02.05 - Can't turn on lights

    newGhost.traits.passive.cantTurnOnLights = ghostName === ghostNames.mare;

            // 03.02.06 - More light-breaking ghost events

    newGhost.traits.passive.moreLightBreakingGhostEvents = ghostName === ghostNames.mare;

            // 03.02.07 - More events in shadow form

    newGhost.traits.passive.moreEventsInShadowForm = ghostName === ghostNames.shade;

            // 03.02.08 - More air ball ghost events

    newGhost.traits.passive.moreAirballGhostEvents = ghostName === ghostNames.shade;

            // 03.02.09 - No activity and hunts if player is in same room

    newGhost.traits.passive.noActivityIfPlayerIsInRoom = ghostName === ghostNames.shade;

            // 03.02.10 - No airball ghost events

    newGhost.traits.passive.noAirballGhostEvents = ghostName === ghostNames.oni;

            // 03.02.11 - Ghost events remove twice as much sanity

    newGhost.traits.passive.eventsDrainDoubleSanity = ghostName === ghostNames.oni;

            // 03.02.12 - Turns breaker off more often

    newGhost.traits.passive.turnsBreakerOffMoreOften = ghostName === ghostNames.hantu;

            // 03.02.13 - Goryo DOTS (only visible through camera and if no player is in room)

    newGhost.traits.passive.goryoDOTS = ghostName === ghostNames.goryo;

            // 03.02.14 - Won't change favourite room

    newGhost.traits.passive.wontChangeGhostRoom = ghostName === ghostNames.goryo;

            // 03.02.15 - Firelights act as crucifixes

    newGhost.traits.passive.fireActsAsCrucifix = ghostName === ghostNames.onryo;

            // 03.02.16 - Can leave 6-finger fingerprints

    newGhost.traits.passive.canLeaveSixFingerFingerprints = ghostName === ghostNames.obake;

            // 03.02.17 - Uses target sanity as earliest hunting parameter

    newGhost.traits.passive.useTargetSanityForHuntStart = ghostName === ghostNames.banshee;

            // 03.02.18 - Events cost 15 instead of 10 sanity

    newGhost.traits.passive.eventsDrainFiftyPercentMoreSanity = ghostName === ghostNames.banshee;

        // 03.03 - Player-Based traits (Caused and visible by player)

    newGhost.traits.playerBased = {};

            // 03.03.01 - Disappears when photographed

    newGhost.traits.playerBased.disappearsWhenPhotographed = ghostName === ghostNames.phantom;

            // 03.03.02 - Para-Mic Banshee Scream

    newGhost.traits.playerBased.bansheeScream = ghostName === ghostNames.banshee;

            // 03.03.03 - Stays in room for a while if smudged

    newGhost.traits.playerBased.staysInRoomIfSmudged = ghostName === ghostNames.yurei;

            // 03.03.04 - Increased activity if talking near ghost

    newGhost.traits.playerBased.talkingIncreasesActivity = ghostName === ghostNames.yokai;

            // 03.03.05 - Breathes in spirit box if on top of ghost

    newGhost.traits.playerBased.spiritBoxBreathing = ghostName === ghostNames.deogen;

            // 03.03.06 - Longer disorientaded when smudged during a hunt

    newGhost.traits.playerBased.longerSmudgedIfSmudgedDuringHunt = ghostName === ghostNames.moroi;

    // 04 - Hunting

    newGhost.hunting = {};

        // 04.01 - Hunt prevention

    newGhost.hunting.huntPrevention = {};

            // 04.01.01 - Smudge Safe Time

    newGhost.hunting.huntPrevention.smudgeSafeTime =
        ghostName === ghostNames.spirit ? 180
    :   ghostName === ghostNames.demon  ? 60
    :                                     90;

            // 04.01.03 - Crucifix Range in %

    newGhost.hunting.huntPrevention.relativeCrucifixRange = ghostName === ghostNames.demon ? 1.5 : 1.0;

            // 04.01.04 - Earliest Hunt after stopped by crucifix (retryTime)

    newGhost.hunting.huntPrevention.retryTime = ghostName === ghostNames.demon ? 20 : 25;

        // 04.02 - Earliest Hunt Start

    newGhost.hunting.huntStart = {};

            // 04.02.01 - Earliest "Earliest Hunt"

    newGhost.hunting.huntStart.earliestPercentage =
        ghostName === ghostNames.mare  ? 60
    :   ghostName === ghostNames.shade ? 35
    :   ghostName === ghostNames.demon ? 70
    :   ghostName === ghostNames.yokai ? 80
    :   ghostName === ghostNames.onryo ? 60
    :   ghostName === ghostNames.raiju ? 65
    :   ghostName === ghostNames.thaye ? 75
    :                                    50;

            // 04.02.02 - Latest "Earliest Hunt"

    newGhost.hunting.huntStart.latestPercentage =
        ghostName === ghostNames.mare  ? 40
    :   ghostName === ghostNames.shade ? 35
    :   ghostName === ghostNames.demon ? 70
    :   ghostName === ghostNames.onryo ? 60
    :   ghostName === ghostNames.thaye ? 15
    :                                    50;
    
            // 04.02.03 - Flags, which condition is important for hunt sanity change

    newGhost.hunting.huntStart.basedOn = {};

                // 04.02.03.01 - Based on Age

    newGhost.hunting.huntStart.basedOn.age = ghostName === ghostNames.thaye;

                // 04.02.03.02 - Based on Room Light

    newGhost.hunting.huntStart.basedOn.roomLight = ghostName === ghostNames.mare;

                // 04.02.03.03 - Based on Player Talking

    newGhost.hunting.huntStart.basedOn.playerTalking = ghostName === ghostNames.yokai;

                // 04.02.03.04 - Based on Electronics

    newGhost.hunting.huntStart.basedOn.electronics = ghostName === ghostNames.raiju;

                // 04.02.03.05 - Based on target's sanity instead of average

    newGhost.hunting.huntStart.basedOn.targetSanity = ghostName === ghostNames.banshee;

            // 04.02.04 - Calculate Earliest Hunt based on parameters

    newGhost.hunting.huntStart.ghost = ghostName;  // This is used, such that the ghost name can be used in calculating the earliestHuntStart inside p ic o f o x'es ghosts.js

    /**
     * Calculates the earliest possible hunt of the ghost with given parameters. Ability-Based Hunt-Starts are not included
     * @param {int} ghostAge current age of the ghost [0..10]
     * @param {boolean} isInDarkRoom true, if the ghost is in a dark room
     * @param {boolean} isPlayerTalkingNearby true, if the player is talking near the ghost
     * @param {boolean} isNearElectronicEquipment true, if the ghost is near electronic equipment
     * @returns integer value representing percentage at which the ghost can hunt using the given parameters
     */
    newGhost.hunting.huntStart.calculateEarliestHunt = function(
        ghostAge = 0,
        isInDarkRoom = false, 
        isPlayerTalkingNearby = false, 
        isNearElectronicEquipment = false
    ) {

        switch (this.ghost) {

            case ghostNames.mare:
                return isInDarkRoom ? this.earliestPercentage : this.latestPercentage;

            case ghostNames.yokai:
                return isPlayerTalkingNearby ? this.earliestPercentage : this.latestPercentage;

            case ghostNames.raiju:
                return isNearElectronicEquipment ? this.earliestPercentage : this.latestPercentage;

            case ghostNames.thaye:
                if (ghostAge < 0)
                    return this.earliestPercentage;
                else if (ghostAge > 10)
                    return this.latestPercentage;
                else
                    return this.earliestPercentage - (ghostAge * ((this.earliestPercentage - this.latestPercentage) / 10));

            default:
                return this.earliestPercentage;

        }

    }

        // 04.03 - Hunting Traits

    newGhost.hunting.huntingTraits = {};

            // 04.03.01 - Flicker times

    newGhost.hunting.huntingTraits.flickerTimeRanges = {};

                // 04.03.01.01 - Visible Time period

    newGhost.hunting.huntingTraits.flickerTimeRanges.visible = {};

                    // 04.03.01.01.01 - From

    newGhost.hunting.huntingTraits.flickerTimeRanges.visible.from =
        ghostName === ghostNames.oni    ? 0.0
    :   ghostName === ghostNames.deogen ? 0.2
    :                                     0.1;

                    // 04.03.01.01.02 - To

    newGhost.hunting.huntingTraits.flickerTimeRanges.visible.to =
        ghostName === ghostNames.oni    ? 0.5
    :                                     0.3;

                // 04.03.01.02 - Invisible Time period

    newGhost.hunting.huntingTraits.flickerTimeRanges.invisible = {};

                    // 04.03.01.02.01 - From

    newGhost.hunting.huntingTraits.flickerTimeRanges.invisible.from =
        ghostName === ghostNames.oni        ? 0.0
    :   ghostName === ghostNames.phantom    ? 0.7
    :                                         0.1;
            
                    // 04.03.01.02.02 - To

    newGhost.hunting.huntingTraits.flickerTimeRanges.invisible.to =
        ghostName === ghostNames.oni        ? 0.5
    :   ghostName === ghostNames.deogen     ? 0.4
    :   ghostName === ghostNames.phantom    ? 1.9
    :                                         0.9;
            
            // 04.03.02 - Walking Speed

    newGhost.hunting.huntingTraits.walkingSpeed = {};

                // 04.03.02.01 - Slowest walking speed

    newGhost.hunting.huntingTraits.walkingSpeed.slowest = 
        ghostName === ghostNames.revenant ? 1.1
    :   ghostName === ghostNames.hantu    ? 1.4
    :   ghostName === ghostNames.twins    ? 1.6
    :   ghostName === ghostNames.deogen   ? 0.4
    :   ghostName === ghostNames.thaye    ? 1.0
    :                                       1.6;

                // 04.03.02.02 - Fastest walking speed

    newGhost.hunting.huntingTraits.walkingSpeed.fastest = 
        ghostName === ghostNames.revenant ? 3.0
    :   ghostName === ghostNames.hantu    ? 2.7
    :   ghostName === ghostNames.twins    ? 2.5
    :   ghostName === ghostNames.moroi    ? 3.9
    :   ghostName === ghostNames.deogen   ? 3.0
    :   ghostName === ghostNames.thaye    ? 2.75
    :                                       2.5;
        
                // 04.03.02.03 - what walking speed is based on ("normal ghosts will only have "ifSpotted" === true)
                
    newGhost.hunting.huntingTraits.walkingSpeed.basedOn = {};

                    // 04.03.02.03.01 - if player is spotted

    newGhost.hunting.huntingTraits.walkingSpeed.basedOn.ifSpotted =
        ghostName === ghostNames.hantu  ? false
    :   ghostName === ghostNames.deogen ? false
    :   ghostName === ghostNames.thaye  ? false
    :                                     true;

                    // 04.03.02.03.02 - distance to player

    newGhost.hunting.huntingTraits.walkingSpeed.basedOn.distanceToPlayer =
        ghostName === ghostNames.deogen ? true
    :   ghostName === ghostNames.jinn   ? true
    :                                     false;
            
                    // 04.03.02.03.03 - average sanity

    newGhost.hunting.huntingTraits.walkingSpeed.basedOn.averageSanity =
        ghostName === ghostNames.moroi  ? true
    :                                     false;
        
                    // 04.03.02.03.04 - temperature

    newGhost.hunting.huntingTraits.walkingSpeed.basedOn.temperature =
        ghostName === ghostNames.hantu  ? true
    :                                     false;
    
                    // 04.03.02.03.05 - age

    newGhost.hunting.huntingTraits.walkingSpeed.basedOn.age =
        ghostName === ghostNames.thaye  ? true
    :                                     false;
                
                    // 04.03.02.03.06 - electronics

    newGhost.hunting.huntingTraits.walkingSpeed.basedOn.electronics =
        ghostName === ghostNames.raiju  ? true
    :                                     false;
            
                    // 04.03.02.03.07 - fuse box

    newGhost.hunting.huntingTraits.walkingSpeed.basedOn.fuseBox =
        ghostName === ghostNames.jinn   ? true
    :                                     false;
        
                    // 04.03.02.03.08 - which twin

    newGhost.hunting.huntingTraits.walkingSpeed.basedOn.hasDecoy =
        ghostName === ghostNames.twins   ? true
    :                                      false;
    
                    // 04.03.02.03.09 - slowly reduce spotted timer to 0 (2.7s from max to 0) instead of preserving its value after line of sight is lost

    newGhost.hunting.huntingTraits.walkingSpeed.basedOn.slowlyReduceSpottedTimer =
        ghostName === ghostNames.revenant   ? true
    :                                         false;
                
                // 04.03.02.04 - time needed for reduce spotted timer after losing line of sight (2.7s from max to 0)

    newGhost.hunting.huntingTraits.walkingSpeed.spottedTimerMaxToNoneTime =
        ghostName === ghostNames.revenant   ? 2.7
    :                                         Infinity;
            
                // 04.03.02.05 - only valid if ghost is using ifSpotted. if true, ghost speed will increase slowly if spotted, otherwise instantly to max speed

    newGhost.hunting.huntingTraits.walkingSpeed.spottedSpeedIncreaseSlow =
        ghostName === ghostNames.revenant   ? false
    :                                         true;
        
                // 04.03.02.07 - only valid if ghost is using hasDecoy.

    newGhost.hunting.huntingTraits.walkingSpeed.decoyModifier =
        ghostName === ghostNames.twins   ? 0.9
    :                                      1.0;
    
                // 04.03.02.08 - only valid if ghost is using hasDecoy.

    newGhost.hunting.huntingTraits.walkingSpeed.nonDecoyModifier =
        ghostName === ghostNames.twins   ? 1.1
    :                                      1.0;
            
                // 04.03.02.09 - only valid if ghost is using distanceToPlayer and alwaysSpotsPlayer.

    newGhost.hunting.huntingTraits.walkingSpeed.distanceThreshholdSpeedMax =
        ghostName === ghostNames.deogen   ? 6.0
    :                                       0.0;
        
                // 04.03.02.10 - only valid if ghost is using distanceToPlayer and alwaysSpotsPlayer.

    newGhost.hunting.huntingTraits.walkingSpeed.distanceThreshholdSpeedMin =
    ghostName === ghostNames.deogen   ? 2.5
:                                       0.0;
    
                // 04.03.02.11 - only used if using averageSanity and ifSpotted

    newGhost.hunting.huntingTraits.walkingSpeed.noSanityMinSpeed =
        ghostName === ghostNames.moroi   ? 2.5
    :                                      newGhost.hunting.huntingTraits.walkingSpeed.slowest;

                // 04.03.02.12 - only used if using averageSanity and ifSpotted

    newGhost.hunting.huntingTraits.walkingSpeed.noSanityMaxSpeed =
        ghostName === ghostNames.moroi   ? 3.9
    :                                      newGhost.hunting.huntingTraits.walkingSpeed.fastest;
    
                // 04.03.02.13 - only used if using averageSanity and ifSpotted

    newGhost.hunting.huntingTraits.walkingSpeed.fullSanityMinSpeed =
        ghostName === ghostNames.moroi   ? 1.6
    :                                      newGhost.hunting.huntingTraits.walkingSpeed.slowest;
    
                // 04.03.02.14 - only used if using averageSanity and ifSpotted

    newGhost.hunting.huntingTraits.walkingSpeed.fullSanityMaxSpeed =
        ghostName === ghostNames.moroi   ? 2.5
    :                                      newGhost.hunting.huntingTraits.walkingSpeed.fastest;
                                            
                // 04.03.02.15 - only used if using temperature

    newGhost.hunting.huntingTraits.walkingSpeed.minSpeedTemperature =
        ghostName === ghostNames.hantu   ? 15.0
    :                                      0.0;   

                // 04.03.02.16 - only used if using temperature

    newGhost.hunting.huntingTraits.walkingSpeed.maxSpeedTemperature =
        ghostName === ghostNames.hantu   ? 0.0
    :                                      0.0;
                                              
                // 04.03.02.16 - only used if using fuseBox

    newGhost.hunting.huntingTraits.walkingSpeed.fuseboxSpeedMinDistance =
        ghostName === ghostNames.jinn    ? 3.0
    :                                      0.0;
                                          
                // 04.03.02.17 - Calculate walking speed

    newGhost.hunting.huntingTraits.walkingSpeed.ghost = ghostName; // Used for calculations

    /**
     * Calculates the ghost's walking speed according to the given parameters and the current ghost
     * @param {float} distanceToPlayer distance to player in meters
     * @param {float} averageSanity average sanity
     * @param {float} temperature Temperature of the ghost's current room
     * @param {int} ghostAge age of the ghost
     * @param {float} spottedTime seconds, how long the ghost has spotted the player
     * @param {boolean} hasLineOfSightToPlayer true, if ghost has currently line of sight to a player
     * @param {boolean} isNearElectronic true, if ghost is near electronic
     * @param {boolean} isFuseBoxOn true, if fuse box is on
     * @param {boolean} isDecoy true, if it's the mimic's decoy hunting
     * @param {int} roundToNthDigit rounds to the nth digit after the decimal point
     * @returns ghost speed in m/s for the current ghost
     */
    newGhost.hunting.huntingTraits.walkingSpeed.calculateSpeed = function (
        distanceToPlayer = 0.0, 
        averageSanity = 100, 
        temperature = 20, 
        ghostAge = 0, 
        spottedTime = 0,
        hasLineOfSightToPlayer = false,
        isNearElectronic = false, 
        isFuseBoxOn = false, 
        isDecoy = false,
        roundToNthDigit = 2
    ) {

        // Spotted time interpolation
        let spottedInterpolationPercentage = spottedTime / 10.0;
        spottedInterpolationPercentage = spottedInterpolationPercentage < 0 ? 0 : spottedInterpolationPercentage > 1 ? 1 : spottedInterpolationPercentage;
        
        let normalSpeedDifference = this.fastest - this.slowest;
        let normalInterpolatedSpeed = this.slowest + normalSpeedDifference * spottedInterpolationPercentage;

        // Special Jinn Calculations
        let jinnThreshhold = this.fuseboxSpeedMinDistance;

        // Special Hantu Calculations - because faster if lower temperature, minTemp = temp at which Speed is max and vice versa
        let minTemperature = this.maxSpeedTemperature; 
        let maxTemperature = this.minSpeedTemperature;

            // Calculate the "percentage" of the current temperature in the range of [maxTemperature..minTemperature], where maxTemperature means 0% and minTemperature means 100%
            // Clamp to 0% or 100% if outside this range
        let temperatureInterpolationPercentage = 1 - ((temperature - minTemperature) / (maxTemperature - minTemperature));
        temperatureInterpolationPercentage = temperatureInterpolationPercentage < 0 ? 0 : temperatureInterpolationPercentage > 1 ? 1 : temperatureInterpolationPercentage;
        
        // Special Moroi Calculations

            // Calculate the inverted sanity percentage
            // Clamp to 0% or 100% if outside this range
        let sanityInterpolationPercentage = 1 - (averageSanity / 100.0);
        sanityInterpolationPercentage = sanityInterpolationPercentage < 0 ? 0 : sanityInterpolationPercentage > 1 ? 1 : sanityInterpolationPercentage;

        let moroiFullSanitySlowest = this.fullSanityMinSpeed;
        let moroiFullSanityFastest = this.fullSanityMaxSpeed;
        let moroiNoSanitySlowest = this.noSanityMinSpeed;
        let moroiNoSanityFastest = this.noSanityMaxSpeed;

        let moroiCurrentSanitySlowest = moroiFullSanitySlowest + (moroiNoSanitySlowest - moroiFullSanitySlowest) * sanityInterpolationPercentage;
        let moroiCurrentSanityFastest = moroiFullSanityFastest + (moroiNoSanityFastest - moroiFullSanityFastest) * sanityInterpolationPercentage;

        let moroiCurrentSanityDifference = moroiCurrentSanityFastest - moroiCurrentSanitySlowest;
        let moroiCurrentSanitySpeed = moroiCurrentSanitySlowest + moroiCurrentSanityDifference * spottedInterpolationPercentage;

        // Special Deogen Calculations
        // sadly, confusing naming because of switched meaning of min and max for distance and speed. let me explain
        let deogenSlowestDistanceMax            // the MaxDistance the deogen can be away for it to be at its Slowest
            = this.distanceThreshholdSpeedMin;  // the distanceThreshhold at which the ghost speed is at its Minimum
        let deogenFastestDistanceMin            // the MinDistance the deogen can be away for it to be at its Fastest
            = this.distanceThreshholdSpeedMax;  // the distanceThreshhold at which the ghost speed is at its Maximum

            // Calculate the percentage of the speed difference for the deogen (slowest if nearest)
            // Clamp to 0% or 100% if outside this range
        let speedInterpolationPercentage = 1 - ((distanceToPlayer - deogenFastestDistanceMin)/(deogenSlowestDistanceMax - deogenFastestDistanceMin));
        speedInterpolationPercentage = speedInterpolationPercentage < 0 ? 0 : speedInterpolationPercentage > 1 ? 1 : speedInterpolationPercentage;

        // Special Thaye Calculations
            // Clamp to 0% or 100% if outside this range
        let ageInterpolationPercentage = ghostAge / 10.0;
        speedInterpolationPercentage = speedInterpolationPercentage < 0 ? 0 : speedInterpolationPercentage > 1 ? 1 : speedInterpolationPercentage;

        let returnValue = normalInterpolatedSpeed;

        // Calculate actual speed and return

        switch (this.ghost) {

            case ghostNames.jinn:
                if (isFuseBoxOn && distanceToPlayer >= jinnThreshhold && hasLineOfSightToPlayer)
                    returnValue = this.fastest;
                else
                    returnValue = normalInterpolatedSpeed;
                break;

            case ghostNames.hantu:
                returnValue = this.slowest + (this.fastest - this.slowest) * temperatureInterpolationPercentage;
                break;

            case ghostNames.twins:
                if (isDecoy)
                    returnValue = normalInterpolatedSpeed * this.decoyModifier;
                else
                    returnValue = normalInterpolatedSpeed * this.nonDecoyModifier;
                break;

            case ghostNames.raiju:
                returnValue = isNearElectronic ? this.fastest : normalInterpolatedSpeed;
                break;

            case ghostNames.moroi:
                returnValue = moroiCurrentSanitySpeed;
                break;

            case ghostNames.deogen:
                returnValue = this.slowest + normalSpeedDifference * speedInterpolationPercentage;
                break;

            case ghostNames.thaye:
                returnValue = this.fastest - normalSpeedDifference * ageInterpolationPercentage;
                break;

            case ghostNames.revenant:
                returnValue = hasLineOfSightToPlayer ? this.fastest : normalInterpolatedSpeed;
                break;

        }

        return roundN(returnValue, roundToNthDigit);

    }

            // 04.03.03 - Ranges

    newGhost.hunting.huntingTraits.ranges = {};

                // 04.03.03.01 - Audio detection range

    newGhost.hunting.huntingTraits.ranges.audioDetectionRange = ghostName === ghostNames.yokai ? 2.5 : 7.5;

                // 04.03.03.02 - Equipment detection range

    newGhost.hunting.huntingTraits.ranges.equipmentDetectionRange = ghostName === ghostNames.yokai ? 2.5 : 9.0;

                // 04.03.03.03 - Range in which the ghost can be heard

    newGhost.hunting.huntingTraits.ranges.canBeHeardWithin = ghostName === ghostNames.myling ? 12 : 20;

                // 04.03.03.04 - Electronics flicker range

    newGhost.hunting.huntingTraits.ranges.electronicFlickerRange = ghostName === ghostName.raiju ? 15 : 10;

            // 04.03.04 - Ghost related special hunting traits

    newGhost.hunting.huntingTraits.special = {};

                // 04.03.04.01 - Item-Throw-Chance at 100% during hunts (chance every half seconds; 1.0 = always, 0.5 = 50% chance every half seconds)

    newGhost.hunting.huntingTraits.special.willAlwaysThrow = ghostName === ghostNames.poltergeist;

                // 04.03.04.02 - Cold breath during hunt

    newGhost.hunting.huntingTraits.special.coldBreath = ghostName === ghostNames.hantu;

                // 04.03.04.03 - Can change ghost model during hunt

    newGhost.hunting.huntingTraits.special.changesGhostModel = ghostName === ghostNames.obake;

                // 04.03.04.04 - Always knows player location

    newGhost.hunting.huntingTraits.special.alwaysDetectsPlayer = ghostName === ghostNames.deogen;

    return newGhost;

}

const Spirit        = createGhost(ghostNames.spirit);
const Wraith        = createGhost(ghostNames.wraith);
const Phantom       = createGhost(ghostNames.phantom);
const Poltergeist   = createGhost(ghostNames.poltergeist);
const Banshee       = createGhost(ghostNames.banshee);
const Jinn          = createGhost(ghostNames.jinn);
const Mare          = createGhost(ghostNames.mare);
const Revenant      = createGhost(ghostNames.revenant);
const Shade         = createGhost(ghostNames.shade);
const Demon         = createGhost(ghostNames.demon);
const Yurei         = createGhost(ghostNames.yurei);
const Oni           = createGhost(ghostNames.oni);
const Yokai         = createGhost(ghostNames.yokai);
const Hantu         = createGhost(ghostNames.hantu);
const Goryo         = createGhost(ghostNames.goryo);
const Myling        = createGhost(ghostNames.myling);
const Onryo         = createGhost(ghostNames.onryo);
const Twins         = createGhost(ghostNames.twins);
const Raiju         = createGhost(ghostNames.raiju);
const Obake         = createGhost(ghostNames.obake);
const Mimic         = createGhost(ghostNames.mimic);
const Moroi         = createGhost(ghostNames.moroi);
const Deogen        = createGhost(ghostNames.deogen);
const Thaye         = createGhost(ghostNames.thaye);

/**
 * Journal containg all the ghost objects containg the ghost informations
 */
const InfoJournal = [
    Spirit,
    Wraith,
    Phantom,
    Poltergeist,
    Banshee,
    Jinn,
    Mare,
    Revenant,
    Shade,
    Demon,
    Yurei,
    Oni,
    Yokai,
    Hantu,
    Goryo,
    Myling,
    Onryo,
    Twins,
    Raiju,
    Obake,
    Mimic,
    Moroi,
    Deogen,
    Thaye
];

function getGhostInfo(ghostName) {
    for (let i = 0; i < InfoJournal.length; i++)
        if (InfoJournal[i].journalInfo.name === ghostName || InfoJournal[i].journalInfo.name === ghostNames[ghostName])
            return InfoJournal[i];
    
    return null;
}