// Builds the ghost journal pages
function buildPages() {

    for (let ghost of InfoJournal)
        buildGhostPage(ghost);

}

// Builds a single ghost journal page
function buildGhostPage(ghost) {
    
    on(
        // Duplicated page template
        generateNewPageElement(), 
        // The ghost's info for the new page
        ghost
    ).execute(
        // Add id to page (journalID)
        setPageID, 
        // Set Ghost name
        setGhostName, 
        // Build link to the wiki
        setWikiLink, 
        // Build Evidence Table
        addEvidenceTable,
        // Build Abilities list
        buildAbilityList,
        // Build Traits list
        buildGhostTraitsList,
        // Build Hunting Conditions section
        buildHuntingConditions,
        // Build Hunting Speed list
        buildHuntingSpeedList,
        //
        buildHuntRangesList,
        //
        buildBlinkingList,
        // Build Hunting Characteristics section
        buildHuntingCharacteristicsList
    ).then(
        // Append page to document body
        callData => BODY.appendChild(callData.inputs[0])
    );
    
    // Build Blinking Behaviour list

    // Build hunting traits section - cold breath, throw chance, ranges during hunts

}

function getTemplateCopy(template) {
    return template.content.cloneNode(true).children[0];
}

function getTemplateCopyOuter(template) {
    return template.content.cloneNode(true);
}

function getPlaceholderElement(parent, placeholderName) {
    return parent.getElementsByClassName(placeholderName)[0];
}

function generateNewPageElement() {
    return getTemplateCopy(TEMPLATE_PAGE);
}

function setPageID(args) {
    args[NEW_PAGE].setAttribute(ATTR_ID, args[GHOST].journalInfo.ID);
}

function setGhostName(args) {
    getPlaceholderElement(args[NEW_PAGE], JS_GHOST_NAME_CLASS).innerHTML = args[GHOST].journalInfo.name;
}

function createWikiLink(ghost) {
    let wikifiedGhostName = ghost.journalInfo.name;
    while (wikifiedGhostName.includes(WIKI_LINK_SPACE))
        wikifiedGhostName = wikifiedGhostName.replace(
            WIKI_LINK_SPACE, 
            WIKI_LINK_SPACE_REPLACEMENT
        );
    return WIKI_LINK_PREFIX + wikifiedGhostName;
}

function setWikiLink(args) {
    getPlaceholderElement(args[NEW_PAGE], JS_WIKI_LINK_CLASS)
        .setAttribute(
            ATTR_HREF, 
            createWikiLink(args[GHOST])
        );
}

function addEvidenceTable(args) {

    let page = args[NEW_PAGE];
    let ghost = args[GHOST];
    let table = getPlaceholderElement(page, JS_EVIDENCE_TABLE_CLASS);

    // Get evidences from ghost
    let totalEvidences = ghost.journalInfo.evidences;

    // Append hidden if not already present
    if (ghost.journalInfo.hiddenEvidence !== null)
        if (!totalEvidences.includes(ghost.journalInfo.hiddenEvidence))
            totalEvidences[totalEvidences.length] = ghost.journalInfo.hiddenEvidence;

    // Append forced if not already present
    if (ghost.journalInfo.forcedEvidence !== null)
        if (!totalEvidences.includes(ghost.journalInfo.forcedEvidence))
            totalEvidences[totalEvidences.length] = ghost.journalInfo.forcedEvidence;

    // Add each evidence
    for (let i = 0; i < totalEvidences.length; i++)
        addEvidenceRow(
            table,                                                   // table
            totalEvidences[i],                                       // evidence name
            totalEvidences[i] !== ghost.journalInfo.hiddenEvidence,  // is hidden
            totalEvidences[i] === ghost.journalInfo.forcedEvidence   // is forced
        );

}

function addEvidenceRow(table, evidenceText, isReal, isForced) {

    let row = getTemplateCopy(TEMPLATE_EVIDENCE_ROW);

    getPlaceholderElement(row, JS_EVIDENCE_NAME_CLASS).innerHTML = evidenceText;
    getPlaceholderElement(row, JS_EVIDENCE_REAL_CLASS).innerHTML = isReal ? CROSS : NO_CROSS;
    getPlaceholderElement(row, JS_EVIDENCE_FORCED_CLASS).innerHTML = isForced ? CROSS : NO_CROSS;

    table.appendChild(row);

}

// because .hasChildNodes() also counts #text as a child node...
function getChildrenCount(element) {
    let count = 0;
    for (let child of element.childNodes)
        if (child.nodeType === Node.ELEMENT_NODE)
            count++;
    return count;
}

function buildList(page, ghostsFlagObject, pathToFlagObject, listClassToInsertInto, textClassIfListIsEmpty) {

    let listElementTemplates = getTemplateCopyOuter(TEMPLATE_LIST_ELEMENT);

    let listElement = getPlaceholderElement(page, listClassToInsertInto);
    let textNone = getPlaceholderElement(page, textClassIfListIsEmpty);

    for (let abilityKey of Object.keys(ghostsFlagObject)) {
        if (!ghostsFlagObject[abilityKey])
            continue;
        let li = listElementTemplates.querySelector(
            ORIGIN_SELECTOR.replace(
                PLACEHOLDER_1, 
                pathToFlagObject
            ).replace(
                PLACEHOLDER_2, 
                abilityKey
            )
        );
        if (li === undefined || li === null)
            continue;
            listElement.appendChild(li);
    }

    if (getChildrenCount(listElement) === 0) {
        listElement.classList.add(HIDDEN_CLASS);
        textNone.classList.remove(HIDDEN_CLASS);
    }

}

function buildAbilityList(args) {
    
    buildList(
        args[NEW_PAGE],
        args[GHOST].abilities,
        JSON_PATH_ABILITIES,
        JS_ABILITIES_LIST_CLASS,
        JS_TEXT_ABILITIES_NONE_CLASS
    );

}

function buildGhostTraitsList(args) {

    buildList(
        args[NEW_PAGE],
        args[GHOST].traits.active,
        JSON_PATH_ACTIVE_TRAITS,
        JS_TRAITS_LIST_ACTIVE_CLASS,
        JS_TRAITS_ACTIVE_NONE_CLASS
    );

    buildList(
        args[NEW_PAGE],
        args[GHOST].traits.passive,
        JSON_PATH_PASSIVE_TRAITS,
        JS_TRAITS_LIST_PASSIVE_CLASS,
        JS_TRAITS_PASSIVE_NONE_CLASS
    );

    buildList(
        args[NEW_PAGE],
        args[GHOST].traits.playerBased,
        JSON_PATH_PLAYERBASED_TRAITS,
        JS_TRAITS_LIST_PLAYERBASED_CLASS,
        JS_TRAITS_PLAYERBASED_NONE_CLASS
    );

}

function getTemplateListElement(elementClass) {
    let template = getTemplateCopyOuter(TEMPLATE_LIST_ELEMENT);
    return template.querySelector(CLASS_SELECTOR + elementClass);
}

function buildVariableLineElement(lineClass, ...params) {
    let line = getTemplateListElement(lineClass);
    // each pair of parameters corresponds to one variable element
    // in the row. parameter 1 is the class of the variable element,
    // parameter 2 is the value to be inserted
    for (let i = 0; i < params.length; i+=2) {
        let span = getPlaceholderElement(
            line,
            params[i]
        );
        // Using booleans instead of parameter values, you can
        // toggle the visibility of an object
        if (typeof params[i+1] === BOOLEAN)
            if (params[i+1])
                span.classList.remove(HIDDEN_CLASS);
            else
                span.classList.add(HIDDEN_CLASS);
        else
            span.innerHTML = params[i+1];
    }
    
    return line;
}

function buildHuntingConditions(args) {
    let page = args[NEW_PAGE];
    let ghost = args[GHOST];

    let huntingConditionsAny = getPlaceholderElement(page, JS_HUNTING_CONDITIONS_LIST_ANY_CLASS);
    let huntingConditionsAll = getPlaceholderElement(page, JS_HUNTING_CONDITIONS_LIST_ALL_CLASS);
    let huntingConditionsSpecial = getPlaceholderElement(page, JS_HUNTING_CONDITIONS_LIST_SPECIAL_CLASS);
    let huntingConditionsPrevent = getPlaceholderElement(page, JS_HUNTING_CONDITIONS_LIST_PREVENT_CLASS);

    // Any: Ghost dependent sanity-threshholds
    //if (ghost.hunting.huntStart.latestPercentage === ghost.hunting.huntStart.earliestPercentage)
        huntingConditionsAny.appendChild(
            buildVariableLineElement(
                JS_HUNT_SANITY_CLASS,
                JS_HUNT_SANITY_AVERAGE_CLASS,
                !ghost.hunting.huntStart.basedOn.targetSanity,
                JS_HUNT_SANITY_TARGET_CLASS,
                ghost.hunting.huntStart.basedOn.targetSanity,
                JS_HUNT_SANITY_PERCENTAGE_CLASS,
                ghost.hunting.huntStart.latestPercentage
            )
        );
    if (ghost.hunting.huntStart.basedOn.age)
        huntingConditionsAny.appendChild(
            buildVariableLineElement(
                JS_HUNT_AGE_CLASS,
                JS_HUNT_AGE_0_CLASS,
                ghost.hunting.huntStart.calculateEarliestHunt(0)
            )
        );
    if (ghost.hunting.huntStart.basedOn.electronics)
        huntingConditionsAny.appendChild(
            buildVariableLineElement(
                JS_HUNT_ELECTRONICS_CLASS,
                JS_HUNT_ELECTRONICS_NEAR_CLASS,
                ghost.hunting.huntStart.calculateEarliestHunt(0, false, false, true)
            )
        );
    if (ghost.hunting.huntStart.basedOn.playerTalking)
        huntingConditionsAny.appendChild(
            buildVariableLineElement(
                JS_HUNT_TALKING_CLASS,
                JS_HUNT_TALKING_PERCENTAGE_CLASS,
                ghost.hunting.huntStart.calculateEarliestHunt(0, false, true)
            )
        );
    if (ghost.hunting.huntStart.basedOn.roomLight)
        huntingConditionsAny.appendChild(
            buildVariableLineElement(
                JS_HUNT_ROOMLIGHT_CLASS,
                JS_HUNT_ROOMLIGHT_OFF_CLASS,
                ghost.hunting.huntStart.calculateEarliestHunt(0, true)
            )
        );

    // All: Not smudged, last hunt ended at least x seconds ago, outside the x-times range of a crucifix/firelight, player not in room for shade
    huntingConditionsAll.appendChild(
        buildVariableLineElement(
            JS_SMUDGE_SAFE_TIME_CLASS,
            JS_SMUDGE_SAFE_TIME_S_CLASS,
            ghost.hunting.huntPrevention.smudgeSafeTime
        )
    );
    huntingConditionsAll.appendChild(
        buildVariableLineElement(
            JS_HUNT_COOLDOWN_CLASS,
            JS_HUNT_COOLDOWN_S_CLASS,
            ghost.hunting.huntPrevention.retryTime
        )
    );
    if (ghost.traits.passive.noActivityIfPlayerIsInRoom) 
        huntingConditionsAll.appendChild(getTemplateListElement(JS_PLAYER_NOT_IN_ROOM_CLASS));
    
    // Special: Cursed hunt, ability hunt
    huntingConditionsSpecial.appendChild(getTemplateListElement(JS_CURSED_POSSESION_HUNT_CLASS));
    if (ghost.abilities.demonHunt)
        huntingConditionsSpecial.appendChild(getTemplateListElement(JS_DEMON_HUNT_CLASS));
    if (ghost.traits.active.huntsIfThreeFiresAreBlownOut)
        huntingConditionsSpecial.appendChild(getTemplateListElement(JS_ONRYO_3RD_FIRE_HUNT_CLASS));
    if (ghost.traits.active.canHuntIfFireIsBlownOut)
        huntingConditionsSpecial.appendChild(getTemplateListElement(JS_ONRYO_POST_3RD_FIRE_HUNT_CLASS));

    // Hunt prevented if; withing crucifix range, firelight (onryo)
    huntingConditionsPrevent.appendChild(
        buildVariableLineElement(
            JS_CRUCIFIX_RANGE_CLASS,
            JS_CRUCIFIX_RANGE_S_CLASS,
            ghost.hunting.huntPrevention.relativeCrucifixRange
        )
    );
    if (ghost.traits.passive.fireActsAsCrucifix)
        huntingConditionsPrevent.appendChild(getTemplateListElement(JS_FIRE_PREVENT_HUNT_CLASS));

} 

function countTrueValuesInObject(obj) {
    return Object.entries(obj).reduce((accumulator, currentValue) => (currentValue[1] === true ? 1 : 0) + accumulator, 0);
}

function buildWalkingSpeedList(speedListElement, ghost) {
    speedListElement.appendChild(
        buildVariableLineElement(
            JS_WALKING_SPEED_NORMAL_CLASS,
            JS_WALKING_SPEED_NORMAL_VALUE_CLASS,
            ghost.hunting.huntingTraits.walkingSpeed.slowest
        )
    );
    speedListElement.appendChild(
        buildVariableLineElement(
            JS_WALKING_SPEED_SPOTTED_CLASS,
            JS_WALKING_SPEED_SPOTTED_VALUE_CLASS,
            ghost.hunting.huntingTraits.walkingSpeed.fastest
        )
    );
}

function buildNormalSpeedList(speedListElement, ghost) {
    buildWalkingSpeedList(
        speedListElement,
        ghost
    );
    speedListElement.appendChild(getTemplateListElement(JS_SPOTTED_INCREASE_STEADILY_CLASS));
    speedListElement.appendChild(getTemplateListElement(JS_SPOTTED_DECREASE_NONE_CLASS));
}

function buildHuntingSpeedList(args) {
    let page = args[NEW_PAGE];
    let ghost = args[GHOST];

    let walkingSpeedList = getPlaceholderElement(page, JS_HUNTING_SPEED_LIST_CLASS);

    // "Normal ghosts"
    if (countTrueValuesInObject(ghost.hunting.huntingTraits.walkingSpeed.basedOn) === 1 && ghost.hunting.huntingTraits.walkingSpeed.basedOn.ifSpotted) {
        buildNormalSpeedList(walkingSpeedList, ghost);
        return;
    }

    // Special ghosts that have one none-modifying min/max speed value pair
    if (![
        ghostNames.moroi, 
        ghostNames.thaye, 
        ghostNames.deogen,
        ghostNames.hantu
    ].includes(ghost.journalInfo.name)) {
        buildWalkingSpeedList(
            walkingSpeedList,
            ghost
        );
    }

    // Special ghosts whose speed will increase slowly/rapidly
    if (ghost.hunting.huntingTraits.walkingSpeed.basedOn.ifSpotted) {

        if (ghost.hunting.huntingTraits.walkingSpeed.spottedSpeedIncreaseSlow)
            walkingSpeedList.appendChild(getTemplateListElement(JS_SPOTTED_INCREASE_STEADILY_CLASS));
        else
            walkingSpeedList.appendChild(getTemplateListElement(JS_SPOTTED_INCREASE_INSTANT_CLASS));

        if (ghost.hunting.huntingTraits.walkingSpeed.basedOn.slowlyReduceSpottedTimer)
            walkingSpeedList.appendChild(getTemplateListElement(JS_SPOTTED_DECREASE_STEADILY_CLASS));
        else
            walkingSpeedList.appendChild(getTemplateListElement(JS_SPOTTED_DECREASE_NONE_CLASS));

    }
    
    // ghosts with single modifiers

        // Jinn: fuse box max speed; twins speedchange, thaye: slower with age
    if (ghost.hunting.huntingTraits.walkingSpeed.basedOn.age) {
        walkingSpeedList.appendChild(getTemplateListElement(JS_SLOWER_BY_AGE_CLASS));
        walkingSpeedList.appendChild(
            buildVariableLineElement(
                JS_FASTEST_BY_AGE_CLASS,
                JS_FASTEST_BY_AGE_S_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.fastest
            )
        );
        walkingSpeedList.appendChild(
            buildVariableLineElement(
                JS_SLOWEST_BY_AGE_CLASS,
                JS_SLOWEST_BY_AGE_S_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.slowest
            )
        );
    }

    if (ghost.hunting.huntingTraits.walkingSpeed.basedOn.distanceToPlayer && ghost.hunting.huntingTraits.special.alwaysDetectsPlayer) {
        walkingSpeedList.appendChild(getTemplateListElement(JS_SLOWER_IF_CLOSER_CLASS));
        walkingSpeedList.appendChild(
            buildVariableLineElement(
                JS_FAR_AWAY_SPEED_CLASS,
                JS_FAR_AWAY_SPEED_DIST_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.distanceThreshholdSpeedMax,
                JS_FAR_AWAY_SPEED_VAL_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.fastest
            )
        );walkingSpeedList.appendChild(
            buildVariableLineElement(
                JS_CLOSE_SPEED_CLASS,
                JS_CLOSE_SPEED_DIST_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.distanceThreshholdSpeedMin,
                JS_CLOSE_SPEED_VAL_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.slowest
            )
        );
    }

    if (ghost.hunting.huntingTraits.walkingSpeed.basedOn.averageSanity) {
        walkingSpeedList.appendChild(getTemplateListElement(JS_LOWER_SANITY_FASTER_CLASS));
        walkingSpeedList.appendChild(
            buildVariableLineElement(
                JS_MAX_SANITY_RANGE_CLASS,
                JS_MAX_SANITY_RANGE_LOWER_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.fullSanityMinSpeed,
                JS_MAX_SANITY_RANGE_UPPER_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.fullSanityMaxSpeed
            )
        );
        walkingSpeedList.appendChild(
            buildVariableLineElement(
                JS_NO_SANITY_RANGE_CLASS,
                JS_NO_SANITY_RANGE_LOWER_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.noSanityMinSpeed,
                JS_NO_SANITY_RANGE_UPPER_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.noSanityMaxSpeed
            )
        );
    }

    if (ghost.hunting.huntingTraits.walkingSpeed.basedOn.electronics)
        walkingSpeedList.appendChild(getTemplateListElement(JS_FULLSPEED_ELECTRO_CLASS));

    if (ghost.hunting.huntingTraits.walkingSpeed.basedOn.hasDecoy)
        walkingSpeedList.appendChild(getTemplateListElement(JS_DECOY_DEPENDANT_SPEED_CLASS));

    if (ghost.hunting.huntingTraits.walkingSpeed.basedOn.temperature) {
        walkingSpeedList.appendChild(getTemplateListElement(JS_COLD_DEPENDANT_SPEED_CLASS));
        walkingSpeedList.appendChild(
            buildVariableLineElement(
                JS_WARM_SPEED_CLASS,
                JS_WARM_SPEED_TEMP_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.minSpeedTemperature,
                JS_WARM_SPEED_VAL_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.slowest
            )
        );
        walkingSpeedList.appendChild(
            buildVariableLineElement(
                JS_COLD_SPEED_CLASS,
                JS_COLD_SPEED_TEMP_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.maxSpeedTemperature,
                JS_COLD_SPEED_VAL_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.fastest
            )
        );
    }

    if (ghost.hunting.huntingTraits.walkingSpeed.basedOn.fuseBox)
        walkingSpeedList.appendChild(
            buildVariableLineElement(
                JS_BREAKER_SPEED_CLASS,
                JS_BREAKER_SPEED_DIST_CLASS,
                ghost.hunting.huntingTraits.walkingSpeed.fuseboxSpeedMinDistance
            )
        );

}

function buildHuntingCharacteristicsList(args) {
    buildList(
        args[NEW_PAGE],
        args[GHOST].hunting.huntingTraits.special,
        JSON_PATH_HUNTING_SPECIAL,
        JS_HUNTING_CHARACTERISTICS_LIST_CLASS,
        JS_HUNTING_CHARACTERISTICS_NONE_CLASS
    );
}

function buildHuntRangesList(args) {
    let page = args[NEW_PAGE];
    let ghost = args[GHOST];
    let rangesList = getPlaceholderElement(page, JS_RANGES_LIST_CLASS);

    rangesList.appendChild(
        buildVariableLineElement(
            JS_RANGE_DETECT_EQUIP_CLASS,
            JS_RANGE_DETECT_EQUIP_VAL_CLASS,
            ghost.hunting.huntingTraits.ranges.equipmentDetectionRange
        )
    );
    rangesList.appendChild(
        buildVariableLineElement(
            JS_RANGE_DETECT_VOICE_CLASS,
            JS_RANGE_DETECT_VOICE_VAL_CLASS,
            ghost.hunting.huntingTraits.ranges.audioDetectionRange
        )
    );
    rangesList.appendChild(
        buildVariableLineElement(
            JS_RANGE_FOOTSTEPS_CLASS,
            JS_RANGE_FOOTSTEPS_VAL_CLASS,
            ghost.hunting.huntingTraits.ranges.canBeHeardWithin
        )
    );
    rangesList.appendChild(
        buildVariableLineElement(
            JS_RANGE_GLITCHING_CLASS,
            JS_RANGE_GLITCHING_VAL_CLASS,
            ghost.hunting.huntingTraits.ranges.electronicFlickerRange
        )
    );
}

function buildBlinkingList(args) {
    let page = args[NEW_PAGE];
    let ghost = args[GHOST];
    let blinkingList = getPlaceholderElement(page, JS_BLINKING_BEHAVIOUR_LIST_CLASS);

    blinkingList.appendChild(
        buildVariableLineElement(
            JS_BLINK_VISIBLE_CLASS,
            JS_BLINK_VISIBLE_MIN_CLASS,
            ghost.hunting.huntingTraits.flickerTimeRanges.visible.from,
            JS_BLINK_VISIBLE_MAX_CLASS,
            ghost.hunting.huntingTraits.flickerTimeRanges.visible.to
        )
    );
    blinkingList.appendChild(
        buildVariableLineElement(
            JS_BLINK_INVISIBLE_CLASS,
            JS_BLINK_INVISIBLE_MIN_CLASS,
            ghost.hunting.huntingTraits.flickerTimeRanges.invisible.from,
            JS_BLINK_INVISIBLE_MAX_CLASS,
            ghost.hunting.huntingTraits.flickerTimeRanges.invisible.to
        )
    );
}

// Initialize pages
buildPages();
initPageElements("right-page", "hidden", "00", "openable-page");