var openedPageClassName;
var hiddenPageClassName;
var defaultPageID;
var openablePageClassName;

function getCurrentlyOpenPage() {
    let openPages = document.querySelectorAll("." + openablePageClassName + "." + openedPageClassName);

    // Error -> Open default page
    if (openPages.length > 1) {
        openPage(defaultPageID);
        return;
    }

    return openPages[0];
}

function getPageById(pageID) {
    return document.getElementById(pageID);
}

/**
 * Initializes document variables about page-toggling.
 * @param {string} openedPage ClassName of the page that is the currently open page
 * @param {string} hiddenClassName ClassName of the class that hides pages
 * @param {string} defaultPage ID of the default page - used for error handling: Sets current page to this page if an error occurs 
 * @param {string} openaplePageClass ClassName of the divs that can be opened as a page
 */
function initPageElements(openedPage = "right-page", hiddenClassName = "hidden", defaultPage = "00", openablePageClass = "openable-page") {
    openedPageClassName = openedPage;
    hiddenPageClassName = hiddenClassName;
    defaultPageID = defaultPage;
    openablePageClassName = openablePageClass;
}

function openPage(pageID) {
    let openedPage = getCurrentlyOpenPage();

    if (openedPage !== undefined) {
        openedPage.classList.remove(openedPageClassName);
        openedPage.classList.add(hiddenPageClassName);
    }

    let newOpenPage = getPageById(pageID);

    if (newOpenPage === null) {
        if (pageID !== defaultPageID)
            openPage(defaultPageID);
        return;
    }

    newOpenPage.classList.remove(hiddenPageClassName);
    newOpenPage.classList.add(openedPageClassName);
}