var deletionReasons = [];

function decreaseEvidence() {
    let evidenceCountElem = document.getElementById("evidence-count");
    let evidenceCount = evidenceCountElem.innerHTML

    if (isNaN(evidenceCount)) {
        console.error(
            "function 'decreaseEvidence', page '" + window.location.href + "': Evidence-Count is NaN"
        );
        alert("An error occured. Please reload the page!");
        return;
    }

    evidenceCount = parseInt(evidenceCount);

    if (evidenceCount > 0)
        evidenceCount--;

    evidenceCountElem.innerHTML = evidenceCount;

}

function increaseEvidence() {
    let evidenceCountElem = document.getElementById("evidence-count");
    let evidenceCount = evidenceCountElem.innerHTML

    if (isNaN(evidenceCount)) {
        console.error(
            "function 'increaseEvidence', page '" + window.location.href + "': Evidence-Count is NaN"
        );
        alert("An error occured. Please reload the page!");
        return;
    }

    evidenceCount = parseInt(evidenceCount);

    if (evidenceCount < 3)
        evidenceCount++;

    evidenceCountElem.innerHTML = evidenceCount;

}

/* Event: Anything clicked: Calculate ghost */
document.addEventListener("click", calculateGhost);

/**
 * the goal of this function is to calculate the ghosts x, where X OF the ghosts are
 * possible for each given evidences. 
 * @param {event} e Will not be used
 * @returns nothing
 */
function calculateGhost(e) {

    deletionReasons = [];
    let evidenceCount = parseInt(document.getElementById("evidence-count").innerHTML);

    ungrayAllEvidences();
    if (evidenceCount === 0)
        grayOutAllEvidencesButOrbs();

    // Get count of selected evidences
    let selectedEvidencesCount = Object.keys(CheckBoxes).filter(v => CheckBoxes[v].isSelected && Object.keys(evidences).includes(v)).length;
    let excludedEvidences = Object.keys(CheckBoxes).filter(v => CheckBoxes[v].isDiscarded && Object.keys(evidences).includes(v));
    let totalExcludedEvidencesCount = excludedEvidences.length;
    let notExcludedEvidences = Object.keys(CheckBoxes).filter(v => !CheckBoxes[v].isDiscarded && Object.keys(evidences).includes(v));

    // Gray every Selectable (every Ghost)
    for (const key of Object.keys(Circleables))
        Circleables[key].element.setAttribute("grayed-out", null);  
    
    // All ghosts are possible at the beginning
    let possibleGhosts = Object.keys(ghostNames);

    // Check evidences
    if (evidenceCount > 0)
        for (const eKey of Object.keys(evidences)) {
            let evidence = evidences[eKey];  
            let evidenceCheckBox = CheckBoxes[eKey];

            // evidence is selected. remove ghosts that dont have these evidences
            if (evidenceCheckBox.isSelected) 
                for (const gKey of possibleGhosts) {
                    let ghostName = gKey;
                    let ghostInfo = getGhostInfo(ghostName);

                    if (ghostInfo.journalInfo.hiddenEvidence === evidence)
                        continue;

                    if (ghostInfo.journalInfo.evidences.includes(evidence))
                        continue;

                    possibleGhosts = possibleGhosts.filter(value => value != ghostName);
                    deletionReasons[ghostName] = "Does not show evidence '" + evidence + "'";
                }

            // evidence is discarded. remove ghosts that have these evidences - only if all 3 evidences are shown
            if (evidenceCount === 3 && evidenceCheckBox.isDiscarded) 
                for (const gKey of possibleGhosts) {
                    let ghostName = gKey;
                    let ghostInfo = getGhostInfo(ghostName);

                    if (ghostInfo.journalInfo.hiddenEvidence === evidence 
                    ||  ghostInfo.journalInfo.evidences.includes(evidence)) {
                        possibleGhosts = possibleGhosts.filter(value => value != ghostName);
                        deletionReasons[ghostName] = "Has evidence '" + evidence + "' but it is not present";
                    }

                }
        }

    // If evidences are < ALL_EVIDENCES and > NO_EVIDENCES and the player selected all possible evidences, then discard ghosts if their hidden evidence is not present
    if (selectedEvidencesCount > 0 && selectedEvidencesCount < 3 && selectedEvidencesCount === evidenceCount)
        for (const gKey of possibleGhosts) {
            let ghostName = gKey;
            let ghostInfo = getGhostInfo(ghostName);

            if (ghostInfo.journalInfo.forcedEvidence === null)
                continue;

            if (ghostInfo.journalInfo.forcedEvidence === ghostInfo.journalInfo.hiddenEvidence)
                continue;

            let forcedEvidenceKey = Object.keys(evidences).find(key => evidences[key] === ghostInfo.journalInfo.forcedEvidence);
            if (forcedEvidenceKey === undefined) {
                console.error(
                    "function 'calculateGhost', page '" + window.location.href + "': forced evidence '" + ghostInfo.journalInfo.forcedEvidence + "' was not null, but could not be resolved to key"
                );
                alert("An error occured. Please reload the page!");
                return;
            }

            if (CheckBoxes[forcedEvidenceKey].isSelected === false) {
                possibleGhosts = possibleGhosts.filter(value => value != ghostName);
                deletionReasons[ghostName] = "Does not show forced evidence '" + evidences[forcedEvidenceKey] + "' although possible number of evidences are selected";
            }

        }

    // If player excluded forced evidence, rule out that ghost
    if (evidenceCount > 0)
        for (const gKey of possibleGhosts) {
            let ghostName = gKey;
            let ghostInfo = getGhostInfo(ghostName);

            if (ghostInfo.journalInfo.forcedEvidence === null)
                continue;

            let forcedEvidenceKey = Object.keys(evidences).find(key => evidences[key] === ghostInfo.journalInfo.forcedEvidence);
            if (forcedEvidenceKey === undefined) {
                console.error(
                    "function 'calculateGhost', page '" + window.location.href + "': forced evidence '" + ghostInfo.journalInfo.forcedEvidence + "' was not null, but could not be resolved to key"
                );
                alert("An error occured. Please reload the page!");
                return;
            }

            if (CheckBoxes[forcedEvidenceKey].isDiscarded) {
                possibleGhosts = possibleGhosts.filter(value => value != ghostName);
                deletionReasons[ghostName] = "Does not show forced evidence '" + evidences[forcedEvidenceKey] + "'";
            }

        }

    // Exclude ghosts that do not show at least some of their evidences based on the difficulty setting. E.g., if you have 2 possible evidences and
    // you could rule out two evidences of a ghost, then rule out that ghost
    if (evidenceCount > 0)
        for (const gKey of possibleGhosts) {
            let ghostName = gKey;
            let ghostInfo = getGhostInfo(ghostName);

            let evidenceKeyArr = ghostInfo.journalInfo.evidences.map(v => Object.keys(evidences).find(key => evidences[key] === v));
            if (ghostInfo.journalInfo.hiddenEvidence !== null)
                evidenceKeyArr[evidenceKeyArr.length] = Object.keys(evidences).find(key => evidences[key] === ghostInfo.journalInfo.hiddenEvidence);

            let excludedEvidencesCount = Object.keys(CheckBoxes).filter(v => CheckBoxes[v].isDiscarded && evidenceKeyArr.includes(v)).length;

            if (excludedEvidencesCount > 3 - evidenceCount) {
                possibleGhosts = possibleGhosts.filter(value => value != ghostName);
                deletionReasons[ghostName] = "Does not show minimum amount of evidences. " + excludedEvidencesCount + " of its evidences were ruled out, but " + evidenceCount + " evidences should be visible";
            }

        }

    // If all but X evidences were ruled out and X = evidence count by difficulty setting, then the left over, non-discarded evidences can be seen as selected
    // Rule out ghosts that do not have these evidences
    if (evidenceCount > 0 && Object.keys(evidences).length - totalExcludedEvidencesCount <= evidenceCount)
        for (const gKey of possibleGhosts) {
            let ghostName = gKey;
            let ghostInfo = getGhostInfo(ghostName);

            let evidenceKeyArr = ghostInfo.journalInfo.evidences.map(v => Object.keys(evidences).find(key => evidences[key] === v));
            if (ghostInfo.journalInfo.hiddenEvidence !== null)
                evidenceKeyArr[evidenceKeyArr.length] = Object.keys(evidences).find(key => evidences[key] === ghostInfo.journalInfo.hiddenEvidence);

            for (let i = 0; i < notExcludedEvidences.length; i++)
                if (!evidenceKeyArr.includes(notExcludedEvidences[i])) {
                    possibleGhosts = possibleGhosts.filter(value => value != ghostName);
                    deletionReasons[ghostName] = "Does not have left-over evidence '" + notExcludedEvidences[i] + "'.";
                }

        }

    // Special case: If there are no evidences and ghost orbs are ....
        // a) found: Only the Mimic must be left
        // b) not found: Exclude the Mimic
    if (evidenceCount === 0)
        if (CheckBoxes["ghostorb"].isSelected)
            for (const gKey of possibleGhosts) {
                let ghostName = gKey;
                let ghostInfo = getGhostInfo(ghostName);

                if (ghostInfo.journalInfo.name !== Mimic.journalInfo.name) {
                    possibleGhosts = possibleGhosts.filter(value => value != ghostName);
                    deletionReasons[ghostName] = "Orbs were found on a no-evidence run";
                }
            }
        else if (CheckBoxes["ghostorb"].isDiscarded)
            for (const gKey of possibleGhosts) {
                let ghostName = gKey;
                let ghostInfo = getGhostInfo(ghostName);

                if (ghostInfo.journalInfo.name === Mimic.journalInfo.name) {
                    possibleGhosts = possibleGhosts.filter(value => value != ghostName);
                    deletionReasons[ghostName] = "Orbs were not found on a no-evidence run";
                }
            }

    // Exclude ghosts if more evidences are selected than there are possible
    // Take hidden evidences into consideration
    if (selectedEvidencesCount > evidenceCount)
        for (const gKey of possibleGhosts) {
            let ghostName = gKey;
            let ghostInfo = getGhostInfo(ghostName);

            let hiddenEvidenceKey = Object.keys(evidences).find(key => evidences[key] === ghostInfo.journalInfo.hiddenEvidence);

            let selectedCount = selectedEvidencesCount;

            if (hiddenEvidenceKey !== undefined)
                if (Object.keys(CheckBoxes).filter(v => CheckBoxes[v].isSelected && Object.keys(evidences).includes(v)).includes(hiddenEvidenceKey))
                    selectedCount--;

            if (selectedCount > evidenceCount) {
                possibleGhosts = possibleGhosts.filter(value => value != ghostName);
                deletionReasons[ghostName] = "More evidences selected than possible";
            }
        }

    // Ungray all possible ghosts
    for (const gKey of possibleGhosts)
        Circleables[gKey].element.removeAttribute("grayed-out");

}

// Calls the execution
function f() {
    let ox = ungrayAllEvidences();
}

function grayOutAllEvidencesButOrbs() {
    let evidenceKeys = Object.keys(CheckBoxes).filter(v => Object.keys(evidences).includes(v));

    for (let i = 0; i < evidenceKeys.length; i++) {
        if (evidenceKeys[i] !== "ghostorb")
            CheckBoxes[evidenceKeys[i]].element.setAttribute("grayed-out", null);
    }
}

function ungrayAllEvidences() {
    let evidenceKeys = Object.keys(CheckBoxes).filter(v => Object.keys(evidences).includes(v));

    for (let i = 0; i < evidenceKeys.length; i++)
        CheckBoxes[evidenceKeys[i]].element.removeAttribute("grayed-out");

}