/**
 * Creates syntax of this kind: on(argument1, argument2, ...).execute(function1, function2, ...).then(callData => <evaluatingFunction>);
 * Pass any number of arguements into the on() part. Then, all these arguments will be turned into an array and passed into the functions
 * on the execute() part as input. The inputs, the function names, and the outputs, will be available in the then() portion of the call.
 * then() will receive one object that contains the members "inputs", which is the input array that comes from the on() part, and a
 * "calls" array, which will contain an object with the function name "func" and the outputs of that call in "output".
 * Use then() to do some final evaluation.
 * @param  {...any} args Arguments that will be used for the function calls
 * @returns Function execute(), which will return the function then()
 */
function on(...args) {
    return {
        /**
         * A list of functions, one parameter = one function, that will be called with the given parameters from on()
         * @param  {...function} fList Each parameter is a function that is called in order of the parameter order
         * @returns Object containing the function then()
         */
        execute: function (...fList) {
            let outs = [];
            for (let func of fList) {
                outs[outs.length] = {
                    func: func.name,
                    output: func(args)
                }
            }
            return {
                /**
                 * Evaluation function that will be given an object containing the inputs and the calls, containing the function names and its outputs 
                 * @param {function} f Function for evaluating the call-chain
                 */
                then: function (f) {
                    f({
                        inputs: args,
                        calls: outs
                    });
                }
            }
        }
    };
} 

/**
 * Creates a new p tag
 * @returns HTML "p" element 
 */
function p() {
    let ico = document.createElement("p");
    return ico;
}

/**
 * Rounds a number to a given digit
 * @param {number} number 
 * @param {number} digit 
 * @returns Rounded number
 */
function roundN(number, digit) {
    return Math.round(number * Math.pow(10, digit)) / Math.pow(10, digit);
}